/* Copyright (c) 2016, NVIDIA CORPORATION. All rights reserved.
 * Copyright (c) 2017, Andreas Herten/Forschungszentrum Jülich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of NVIDIA CORPORATION nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "papi.h"
#define PAPI_CALL( call, success )                                            \
{                                                                             \
    int err = call;                                                           \
    if ( success != err)                                                      \
    printf("PAPI error for %s in L%d of %s: %s\n", #call, __LINE__, __FILE__, PAPI_strerror(err)); \
}

// Run with:
// ./poisson2d [NITER [NY [NX]]]
int main(int argc, char** argv)
{
    int ny = 2048;
    int nx = 2048;
    int iter_max = 500;
    const real tol = 1.0e-5;

    if (argc >= 2)
    {
        iter_max = atoi( argv[1] );
    }
    if (argc == 3)
    {
        ny = atoi( argv[2] );
        nx = ny;
    }
    if (argc == 4)
    {
        ny = atoi( argv[2] );
        nx = atoi( argv[3] );
    }

    // Init PAPI
    int EventSet = PAPI_NULL;

    int N_COUNTER = 1;
    const char * counterNames[N_COUNTER];
    counterNames[0] = "PM_VECTOR_FLOP_CMPL";  // TODO: Insert counter for vector floating point operations completed

    // % PAPI: Initialize
    PAPI_CALL( PAPI_library_init(PAPI_VER_CURRENT), PAPI_VER_CURRENT );
    
    // % PAPI: Create event container
    PAPI_CALL( PAPI_create_eventset(&EventSet), PAPI_OK );

    int perr = PAPI_OK;
    for (int i = 0; i < N_COUNTER; i++) {
        // % PAPI: Test availability of counters
        perr = PAPI_query_named_event(counterNames[i]);
        if (perr != PAPI_OK) {
            printf("Event %s can not be found. Exiting.\n", counterNames[i]);
            printf("Error: %s\n", PAPI_strerror(perr));
            exit(1);
        }

        // % PAPI: Add event
        perr = PAPI_add_named_event(EventSet, counterNames[i]);
        if (perr != PAPI_OK ) {
            printf("Event %s can not be added to event set. Probably too many counters in measurement already. Exiting.\n", counterNames[i]);
            printf("Error: %s\n", PAPI_strerror(perr));
            exit(1);
        }
    }

    // % PAPI: Prepare data structure to collect values in
    unsigned long long int pmuvalue[iter_max][2];

    real* restrict const A    = (real*) malloc(nx*ny*sizeof(real));
    real* restrict const Aref = (real*) malloc(nx*ny*sizeof(real));
    real* restrict const Anew = (real*) malloc(nx*ny*sizeof(real));
    real* restrict const rhs  = (real*) malloc(nx*ny*sizeof(real));
    
    // set rhs
    for (int iy = 1; iy < ny-1; iy++)
    {
        for( int ix = 1; ix < nx-1; ix++ )
        {
            const real x = -1.0 + (2.0*ix/(nx-1));
            const real y = -1.0 + (2.0*iy/(ny-1));
            rhs[iy*nx+ix] = expr(-10.0*(x*x + y*y));
        }
    }

    int ix_start = 1;
    int ix_end   = (nx - 1);

    int iy_start = 1;
    int iy_end   = (ny - 1);

    for( int iy = 0; iy < ny; iy++)
    {
        for( int ix = 0; ix < nx; ix++ )
        {
            A[iy*nx+ix] = 0.0;
        }
    }
    
    
    double start = get_time();
    int iter  = 0;
    real error = 1.0;
    
    while ( error > tol && iter < iter_max )
    {  
        // % PAPI: Start collection
        PAPI_CALL( PAPI_start(EventSet), PAPI_OK );
        error = 0.0;

        for (int iy = iy_start; iy < iy_end; iy++)
        {
            for( int ix = ix_start; ix < ix_end; ix++ )
            {
                Anew[iy*nx+ix] = -0.25 * (rhs[iy*nx+ix] - ( A[iy*nx+ix+1] + A[iy*nx+ix-1]
                                                       + A[(iy-1)*nx+ix] + A[(iy+1)*nx+ix] ));
                error = fmaxr( error, fabsr(Anew[iy*nx+ix]-A[iy*nx+ix]));
            }
        }

        // % PAPI: Stop call and add value to array
        PAPI_CALL( PAPI_stop(EventSet, pmuvalue[iter]), PAPI_OK );
        
        for (int iy = iy_start; iy < iy_end; iy++)
        {
            for( int ix = ix_start; ix < ix_end; ix++ )
            {
                A[iy*nx+ix] = Anew[iy*nx+ix];
            }
        }

        //Periodic boundary conditions
        for (int ix = ix_start; ix < ix_end; ix++)
        {
                A[0*nx+ix]      = A[(ny-2)*nx+ix];
                A[(ny-1)*nx+ix] = A[1*nx+ix];
        }
        for (int iy = iy_start; iy < iy_end; iy++)
        {
                A[iy*nx+0]      = A[iy*nx+(nx-2)];
                A[iy*nx+(nx-1)] = A[iy*nx+1];
        }

        iter++;
    }
    double runtime = get_time() - start;
    unsigned long long int counter_tot[N_COUNTER];
    unsigned long long int counter_min[N_COUNTER];
    unsigned long long int counter_max[N_COUNTER];
    for (int i = 0; i < N_COUNTER; i++) {
        counter_tot[i] = 0;
        counter_min[i] = pmuvalue[0][i];  // Init with first value
        counter_max[i] = pmuvalue[0][i];  // Init with first value
    }
    for (int i_iter = 0; i_iter < iter; i_iter++) {
        for (int i_counter = 0; i_counter < N_COUNTER; i_counter++) {
            counter_tot[i_counter] += pmuvalue[i_iter][i_counter];
            counter_min[i_counter]  = (counter_min[i_counter] < pmuvalue[i_iter][i_counter]) ? counter_min[i_counter] : pmuvalue[i_iter][i_counter];
            counter_max[i_counter]  = (counter_max[i_counter] > pmuvalue[i_iter][i_counter]) ? counter_max[i_counter] : pmuvalue[i_iter][i_counter];
        }
    }
    
    printf("iter,ny,nx,Runtime");
    for (int i = 0; i < N_COUNTER; i++) {
        printf(",%s (total),%s (min), %s (max)", counterNames[i], counterNames[i], counterNames[i]);
    }
    printf("\n");

    printf("%d,%d,%d,%.4f", iter_max, ny, nx, runtime);
    for (int i = 0; i < N_COUNTER; i++) {
        printf(",%d,%d,%d", counter_tot[i], counter_min[i], counter_max[i]);
    }
    printf("\n");
    
    free(rhs);
    free(Anew);
    free(Aref);
    free(A);

    return 0;
}
