! Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions
! are met:
!  * Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
!  * Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation and/or other materials provided with the distribution.
!  * Neither the name of NVIDIA CORPORATION nor the names of its
!    contributors may be used to endorse or promote products derived
!    from this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
! EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
! PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
! CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
! OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
! OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

PROGRAM poisson2d
    IMPLICIT NONE
    INTEGER, PARAMETER :: MAX_ITER_MAX_DIGITS = 512
    INTEGER, PARAMETER :: nx = 2048
    INTEGER, PARAMETER :: ny = 2048
    REAL, PARAMETER :: tol = 1.0E-5
    INTEGER :: i,ix, iy, ix_start, ix_end, iy_start, iy_end, iter, iter_max, ierror
    REAL :: x,y, error
    REAL*8 :: runtime_cpu, runtime, start, finish
    LOGICAL, EXTERNAL :: check_results
    LOGICAL :: errors
    REAL, DIMENSION(:,:), ALLOCATABLE :: a, a_ref, a_new, rhs
    CHARACTER(MAX_ITER_MAX_DIGITS) :: iter_max_arg
    
    call getarg(1, iter_max_arg)
    
    IF ( iter_max_arg == '' ) THEN
        iter_max = 500
    ELSE
        read (iter_max_arg, *) iter_max
    ENDIF
    
    ALLOCATE( a(nx,ny) )
    ALLOCATE( a_ref(nx,ny) )
    ALLOCATE( a_new(nx,ny) )
    ALLOCATE( rhs(nx,ny) )
    
    a = 0.0
    a_ref = 0.0
    
    DO iy = 2, ny-1
        DO ix = 2, nx-1
            x = -1.0 + (2.0*ix/(nx-1.0))
            y = -1.0 + (2.0*iy/(ny-1.0))
            rhs(ix,iy) = EXP(-10.0*(x*x+y*y))
        END DO
    END DO
    
    ix_start = 2
    ix_end   = nx-1
    
    iy_start = 2
    iy_end = ny-1
    
    !OpenACC Warm-up
    !$acc parallel loop
    DO iy = 1, ny
        DO ix = 1, nx
            a(ix,iy) = 0.0
        END DO
    END DO
    !$acc end parallel
        

    WRITE(*,"('Jacobi relaxation Calculation: ',I4,' x ',I4,' mesh')") nx,ny
    WRITE(*,*) 'Calculate reference solution and time serial CPU execution.'
    call cpu_time(start)
    CALL poisson2d_serial( nx, ny, iter_max, tol, a_ref, a_new, rhs )
    call cpu_time(finish)
    runtime_cpu = finish-start
    
    WRITE(*,*) 'GPU execution.'
    
    call cpu_time(start)
    iter = 1
    error = 1.0
    DO WHILE ( error > tol .AND. iter <= iter_max )
        error = 0.0
        !$acc parallel loop
        DO iy = iy_start, iy_end
            DO ix = ix_start, ix_end
                a_new(ix,iy) = -0.25 * (rhs(ix,iy) - ( a(ix+1,iy) + a(ix-1,iy) + a(ix,iy-1) + a(ix,iy+1) ))
                error = MAX( error, ABS( a_new(ix,iy) - a(ix,iy) ) )
            END DO
        END DO
        !$acc end parallel
        
        !$acc parallel loop
        DO iy = iy_start, iy_end
            DO ix = ix_start, ix_end
                a(ix,iy) = a_new(ix,iy)
            END DO
        END DO
        !$acc end parallel
        
        !$acc parallel loop
        DO ix = ix_start, ix_end
            a(ix,1) = a(ix,ny-1)
            a(ix,ny) = a(ix,2)
        END DO
        !$acc end parallel
        !$acc parallel loop
        DO iy = iy_start, iy_end
            a(1,iy) = a(nx-1,iy)
            a(nx,iy) = a(2,iy)
        END DO
        !$acc end parallel

        IF ( iter == 1 .OR. MOD( iter, 100 ) == 0 ) THEN
            WRITE(*,"('  ',I4,' ',F8.6)") iter, error
        END IF
        
        iter = iter+1
    END DO
    call cpu_time(finish)
    runtime = finish-start
    
    errors = .FALSE.
    IF ( check_results( ix_start, ix_end, iy_start, iy_end, nx, ny, tol, a, a_ref ) ) THEN
        WRITE(*,"(I4,'x',I4,': 1 CPU: ',F8.4,' s 1 GPU: ',F8.4,' s, speedup: ',F8.2)"), &
              nx,ny,runtime_cpu,runtime,runtime_cpu/runtime
    ELSE
        errors = .TRUE.
    END IF

    DEALLOCATE( rhs )
    DEALLOCATE( a_new )
    DEALLOCATE( a_ref )
    DEALLOCATE( a )
    IF ( errors ) THEN
        STOP -1
    END IF
END PROGRAM poisson2d
